/* $Cambridge: hermes/src/prayer/prayer/banner.c,v 1.1.1.1 2003/04/15 13:00:03 dpc22 Exp $ */
/************************************************
 *    Prayer - a Webmail Interface              *
 ************************************************/

/* Copyright (c) University of Cambridge 2000 - 2002 */
/* See the file NOTICE for conditions of use and distribution. */

#include "prayer_session.h"

/* Method for manipulating icon banner which can appear at the top and
 * bottom of each screen, together with descriptive text
 */

/* banner_create() ******************************************************
 *
 * Create a fresh banner structure
 *  pool: Pool to allocate from
 *
 * Returns: New banner structure 
 ***********************************************************************/

struct banner *banner_create(struct pool *pool)
{
    struct banner *b = pool_alloc(pool, sizeof(struct banner));

    b->first = b->last = NIL;
    b->pool = pool;
    b->full_width = NIL;
    b->use_submit = NIL;

    return (b);
}

/* banner_reset() *******************************************************
 *
 * Reset banner to empty state.
 *         bl: Banner list
 *       pool: Pool to allocate from
 * full_width: Banner should take full width of the display
 *
 ***********************************************************************/

void banner_reset(struct banner *bl, struct pool *pool, BOOL full_width)
{
    bl->pool = pool;
    bl->first = bl->last = NIL;
    bl->full_width = full_width;
    bl->use_submit = NIL;
}

/* ====================================================================== */

/* banner_use_submit() ***************************************************
 *
 * Use submit methods rather than links for form.
 *         bl: Banner list
 * use_submit: T => enable.
 *
 ***********************************************************************/

void banner_use_submit(struct banner *bl, BOOL use_submit)
{
    bl->use_submit = use_submit;
}

/* banner_add() *********************************************************
 *
 * Add icon to banner list
 *     bl: Banner list
 *   text: Descriptive text that should be added under icon and as alt tag
 *   icon: Icon filename
 *   link: Session URL link.
 *
 ***********************************************************************/

void banner_add(struct banner *bl, char *text, char *icon, char *link)
{
    struct pool *pool = bl->pool;
    struct banner_item *new = pool_alloc(pool, sizeof(struct banner_item));

    new->next = NIL;
    new->icon = pool_strdup(pool, icon);
    new->text = pool_strdup(pool, text);
    new->link = NIL;
    new->msgno = 0;

    if (link)
        new->link = pool_strdup(pool, link);

    if (bl->last)
        bl->last = bl->last->next = new;
    else
        bl->first = bl->last = new;
}

/* banner_msg_add() *****************************************************
 *
 * Add icon relating to specific message in current folder to banner list
 *     bl: Banner list
 *   text: Descriptive text that should be added under icon and as alt tag
 *   icon: Icon filename
 *   link: Session URL link.
 *  msgno: Message number
 *
 ***********************************************************************/

void
banner_msg_add(struct banner *bl,
               char *text, char *icon, char *link, unsigned long msgno)
{
    struct pool *pool = bl->pool;
    struct banner_item *new = pool_alloc(pool, sizeof(struct banner_item));

    new->next = NIL;
    new->icon = pool_strdup(pool, icon);
    new->text = pool_strdup(pool, text);
    new->link = NIL;
    new->msgno = msgno;

    if (link)
        new->link = pool_strdup(pool, link);

    if (bl->last)
        bl->last = bl->last->next = new;
    else
        bl->first = bl->last = new;
}

/* ====================================================================== */

/* banner_print_icons() *************************************************
 *
 * Print icon line as HTML table row.
 *      bl: Banner list
 *       b: Target buffer
 * session: Session structure
 ***********************************************************************/

static void
banner_print_icons(struct banner *bl,
                   struct buffer *b, struct session *session)
{
    MAILSTREAM *stream = session->stream;
    struct banner_item *current;
    unsigned long msgno;

    bputs(b, "<tr>");
    for (current = bl->first; current; current = current->next) {
        bputs(b, "<td align=\"center\" valign=\"top\">");
        if (current->link && current->link[0]) {
            /* Active link */
            if ((msgno = current->msgno) > 0)
                html_session_bprintf(session, b, "<a href=\"", "\">",
                                     "%s/%lu/%lu", current->link,
                                     msgno, ml_uid(session, stream,
                                                   msgno));
            else
                html_session_bputs(session, b, "<a href=\"", "\">",
                                   current->link);

            html_icon(session, b, current->icon, current->text);
            bputs(b, "</a>");
        } else                  /* Inactive link */
            html_icon(session, b, current->icon, current->text);
        bputs(b, "</td>" CRLF);
        if (current->next && !bl->full_width)
            bputs(b, "<td></td>" CRLF);
    }
    bputs(b, "</tr>" CRLF);
}

/* ====================================================================== */

/* banner_print_text() **************************************************
 *
 * Print text line as HTML table row.
 *       bl: Banner list
 *        b: Target buffer
 *  session: Session structure
 ***********************************************************************/

static void
banner_print_text(struct banner *bl,
                  struct buffer *b, struct session *session)
{
    MAILSTREAM *stream = session->stream;
    struct banner_item *current;
    unsigned long msgno;
    struct user_agent *user_agent = session->user_agent;

    bputs(b, "<tr>" CRLF);
    for (current = bl->first; current; current = current->next) {
        if (user_agent->use_icons)
            bputs(b, "<td align=\"center\" valign=\"top\">");
        else
            bputs(b, "<td align=\"center\">");
        if (current->link && current->link[0]) {
            /* Active link */
            if ((msgno = current->msgno) > 0)
                html_session_bprintf(session, b, "<a href=\"", "\">",
                                     "%s/%lu/%lu", current->link,
                                     msgno, ml_uid(session, stream,
                                                   msgno));
            else
                html_session_bputs(session, b, "<a href=\"", "\">",
                                   current->link);

            html_quote_string(b, current->text);
            bputs(b, "</a>");
        } else {                /* Inactive link */
            html_quote_string(b, current->text);
        }
        bputs(b, "</td>" CRLF);
        if (current->next && !bl->full_width)
            bputs(b, "<td>|</td>" CRLF);
    }
    bputs(b, "</tr>" CRLF);
}

/* ====================================================================== */

/* banner_print_icons_submit() *******************************************
 *
 * Print icon line as HTML table row.
 *      bl: Banner list
 *       b: Target buffer
 * session: Session structure
 ***********************************************************************/

static void
banner_print_icons_submit(struct banner *bl,
                          struct buffer *b, struct session *session)
{
    struct request *request = session->request;
    struct user_agent *user_agent = request->user_agent;
    char *border = (user_agent->is_netscape4) ? "border=0 " : "";
    struct banner_item *current;
    unsigned long msgno;
    MAILSTREAM *stream = session->stream;

    bputs(b, "<tr>");
    for (current = bl->first; current; current = current->next) {
        bputs(b, "<td align=\"center\" valign=\"top\">");
        if (current->link && current->link[0]) {
            bputs(b, "<input type=image ");

            /* Active link */
            if ((msgno = current->msgno) > 0)
                bprintf(b,
                        "src=\"%s/%s\" %salt=\"%s\" name=\"sub_%s/%lu/%lu\">",
                        session->url_prefix_icons, current->icon, border,
                        current->text, current->link,
                        msgno, ml_uid(session, stream, msgno));
            else
                bprintf(b,
                        "src=\"%s/%s\" %salt=\"%s\" name=\"sub_%s\">",
                        session->url_prefix_icons, current->icon, border,
                        current->text, current->link);
        } else                  /* Inactive link */
            html_icon(session, b, current->icon, current->text);

        bputs(b, "</td>" CRLF);
        if (current->next && !bl->full_width)
            bputs(b, "<td></td>" CRLF);
    }
    bputs(b, "</tr>" CRLF);
}

/* ====================================================================== */

/* banner_print_text_inactive() ******************************************
 *
 * Print text line as inactive HTML table row for annotation
 *       bl: Banner list
 *        b: Target buffer
 *  session: Session structure
 ***********************************************************************/

static void
banner_print_text_inactive(struct banner *bl,
                           struct buffer *b, struct session *session)
{
    struct banner_item *current;
    struct user_agent *user_agent = session->user_agent;

    bputs(b, "<tr>" CRLF);
    for (current = bl->first; current; current = current->next) {
        if (user_agent->use_icons)
            bputs(b, "<td align=\"center\" valign=\"top\">");
        else
            bputs(b, "<td align=\"center\">");

        html_quote_string(b, current->text);
        bputs(b, "</td>" CRLF);
        if (current->next && !bl->full_width)
            bputs(b, "<td>|</td>" CRLF);
    }
    bputs(b, "</tr>" CRLF);
}

/* ====================================================================== */

/* banner_print_text_submit() *******************************************
 *
 * Print text line as HTML table row.
 *       bl: Banner list
 *        b: Target buffer
 *  session: Session structure
 ***********************************************************************/

static void
banner_print_text_submit(struct banner *bl,
                         struct buffer *b, struct session *session)
{
    struct user_agent *user_agent = session->user_agent;
    MAILSTREAM *stream = session->stream;
    struct banner_item *current;
    unsigned long msgno;

    bputs(b, "<tr>" CRLF);
    for (current = bl->first; current; current = current->next) {
        if (user_agent->use_icons)
            bputs(b, "<td align=\"center\" valign=\"top\">");
        else
            bputs(b, "<td align=\"center\">");

        if (current->link && current->link[0]) {
            /* Active link */
            if ((msgno = current->msgno) > 0)
                bprintf(b, ("<input type=\"submit\" value=\"%s\" "
                            "name=\"sub_%s/%lu/%lu\">"),
                        current->text,
                        current->link, msgno, ml_uid(session, stream,
                                                     msgno));
            else
                bprintf(b,
                        "<input type=\"submit\" value=\"%s\" name=\"sub_%s\">",
                        current->text, current->link);
        } else
            /* Inactive link */
            html_quote_string(b, current->text);

        bputs(b, "</td>" CRLF);
        if (current->next && !bl->full_width)
            bputs(b, "<td></td>" CRLF);
    }
    bputs(b, "</tr>" CRLF);
}

/* ====================================================================== */

#if 0
/* banner_print_js_submit() *********************************************
 *
 * Print text line as HTML table row.
 *       bl: Banner list
 *        b: Target buffer
 *  session: Session structure
 ***********************************************************************/

static void
banner_print_js_submit(struct banner *bl,
                       struct buffer *b, struct session *session)
{
    MAILSTREAM *stream = session->stream;
    struct banner_item *current;
    unsigned long msgno;

    bputs(b, "<tr>" CRLF);
    for (current = bl->first; current; current = current->next) {
        bputs(b, "<td align=\"center\" valign=\"top\">");
        if (current->link && current->link[0]) {
            /* Active link */
            if ((msgno = current->msgno) > 0) {
                bprintf(b,
                        "<a onClick=\"form.submit('sub_%s/%lu/%lu'); return true;\" ",
                        current->link, msgno, ml_uid(session, stream,
                                                     msgno));
                html_session_bprintf(session, b, "href=\"", "\">",
                                     "%s/%lu/%lu", current->link, msgno,
                                     ml_uid(session, stream, msgno));
            } else {
                bprintf(b,
                        "<a onClick=\"javascript:form.submit('sub_%s'); return true;\" ",
                        current->link);
                html_session_bputs(session, b, "href=\"", "\">",
                                   current->link);
            }

            html_quote_string(b, current->text);
            bputs(b, "</a>");
        } else {
            /* Inactive link */
            html_quote_string(b, current->text);
        }
        bputs(b, "</td>" CRLF);
    }
    bputs(b, "</tr>" CRLF);
}
#endif

/* ====================================================================== */

/* banner_print() *******************************************************
 *
 * Print banner as pair of HTML rows in a table.
 *      bl: Banner list
 *       b: Target buffer
 * session: Session structure
 ***********************************************************************/

void
banner_print(struct banner *bl, struct buffer *b, struct session *session)
{
    struct user_agent *user_agent = session->user_agent;

    if (bl->full_width)
        bputs(b, "<table width=\"100%\">" CRLF);
    else
        bputs(b, "<table>" CRLF);

    if (bl->use_submit) {
        if (user_agent->use_icons) {
            banner_print_icons_submit(bl, b, session);
            banner_print_text_inactive(bl, b, session);
        } else
            banner_print_text_submit(bl, b, session);
    } else {
        if (user_agent->use_icons) {
            banner_print_icons(bl, b, session);
            banner_print_text_inactive(bl, b, session);
        } else
            banner_print_text(bl, b, session);
    }

    bputs(b, "</table>" CRLF);
}
