/* $Cambridge: hermes/src/prayer/prayer/dircache.c,v 1.1.1.1 2003/04/15 13:00:03 dpc22 Exp $ */
/************************************************
 *    Prayer - a Webmail Interface              *
 ************************************************/

/* Copyright (c) University of Cambridge 2000 - 2002 */
/* See the file NOTICE for conditions of use and distribution. */

#include "prayer_session.h"

/* dircache is a tree of directory listings (dirlist). */

/* ====================================================================== */

/* Static utility routine used by dircache_create() and others */

static struct dircache *dircache_create_internal(struct pool *pool)
{
    struct dircache *result = pool_alloc(pool, sizeof(struct dircache));
    result->pool = pool;
    result->next = NIL;
    result->children = NIL;
    result->name = NIL;
    result->dirlist = NIL;

    return (result);
}

/* dircache_create() ******************************************************
 *
 * Create a new (root-level) directory cache, including its own directory
 * cache.
 *************************************************************************/

struct dircache *dircache_create(void)
{
    return (dircache_create_internal
            (pool_create(DIRCACHE_PREFERRED_POOL_SIZE)));
}

/* dircache_free() *******************************************************
 *
 * Free (pool of) this dircache
 ************************************************************************/

void dircache_free(struct dircache *dc)
{
    pool_free(dc->pool);
}

/* ====================================================================== */

/* Static utility routine for dircache_lookup() */

static struct dircache *dircache_lookup_internal(struct dircache *dc,
                                                 char *name)
{
    struct dircache *current;
    char *next;

    if (!(name && name[0]))
        return (dc);

    if ((next = strchr(name, '/')))
        *next = '\0';

    for (current = dc->children; current; current = current->next) {
        if (!strcmp(current->name, name)) {
            if (next)
                *next++ = '/';
            return (dircache_lookup_internal(current, next));
        }
    }
    if (next)
        *next++ = '/';
    return (NIL);
}

/* dircache_lookup() ******************************************************
 *
 * Look up entry in (root level) directory cache.
 *   session:
 *      name: Name of directory to lookup
 *
 * Returns: relevant directory cache
 *************************************************************************/

struct dircache *dircache_lookup(struct session *session, char *name)
{
    return (dircache_lookup_internal(session->dircache, name));
}

/* ====================================================================== */

/* Static utility routine for dircache_insert */

static struct dircache *dircache_insert_internal(struct pool *p,
                                                 struct dircache *dc,
                                                 char *name)
{
    struct dircache *current;
    char *next;

    if (!(name && name[0]))     /* Shouldn't be possible! */
        return (NIL);

    if ((next = strchr(name, '/')))
        *next = '\0';

    for (current = dc->children; current; current = current->next) {
        if (!strcmp(current->name, name)) {
            if (next)
                *next++ = '/';
            return (dircache_insert_internal(p, current, next));
        }
    }

    while (name) {
        current = dircache_create_internal(dc->pool);
        current->next = dc->children;
        dc->children = current;
        dc = current;
        dc->name = pool_strdup(dc->pool, name);

        if (next)
            *next++ = '/';

        name = next;
        if (name && (next = strchr(name, '/')))
            *next = '\0';
    }

    return (dc);
}

/* dircache_insert() *****************************************************
 *
 * Insert entry into dircache (creates placeholder which can then
 * be populated with a dirlist)
 *    session:
 *       name: Directory name to be inserted into dircache
 *
 * Returns: Ptr to relevant/new dircache entry.
 ************************************************************************/

struct dircache *dircache_insert(struct session *session, char *name)
{
    struct dircache *dc = session->dircache;

    if (!(name && name[0]))     /* Top level special case! */
        return (dc);

    return (dircache_insert_internal(dc->pool, dc, name));
}

/* ====================================================================== */

/* dircache_fetch() ******************************************************
 *
 * Fetch dircache entry for given directory name. (Creates new dircache
 * entry and populates it if necessary).
 *   session:
 *      name: Directory name to lookup
 *
 ************************************************************************/

struct dirlist *dircache_fetch(struct session *session, char *name)
{
    struct dircache *dc;
    struct dirlist *dirlist = NIL;

    if ((dc = dircache_lookup(session, name)) && (dc->dirlist))
        return (dc->dirlist);

    if (!(dirlist = dirlist_populate(session, name)))
        return (NIL);

    if (!(dc = dircache_insert(session, name)))
        return (NIL);

    return (dc->dirlist = dirlist);
}

/* ====================================================================== */

/* (Recursive) Static support routine for dircache_invalidate () */

BOOL dircache_invalidate_internal(struct dircache * dc, char *name)
{
    struct dircache *last, *current;
    char *next;

    if (!(name && name[0]))     /* Shouldn't be possible! */
        return (NIL);

    if ((next = strchr(name, '/')))
        *next = '\0';

    last = NIL;
    for (current = dc->children; current; current = current->next) {
        if (!strcmp(current->name, name)) {
            if (next) {
                *next++ = '/';
                return (dircache_invalidate_internal(current, next));
            } else {
                if (last)
                    last->next = current->next; /* Remove link from middle */
                else
                    dc->children = current->next;       /* Remove link from start */
                return (T);
            }
        }
        last = current;
    }
    if (next)
        *next++ = '/';

    return (NIL);
}

/* dircache_invalidate() *************************************************
 *
 * Invalidate directory cache for specific directory (and subdirectories)
 *   session:
 *      name: Directory tree to invalidate
 *
 * Returns: T if directory was invalidated. NIl => Doesn't exist.
 ************************************************************************/

BOOL dircache_invalidate(struct session * session, char *name)
{
    if (name && name[0])
        return (dircache_invalidate_internal(session->dircache, name));

    /* Invalidate everything */
    dircache_free(session->dircache);

    session->dircache = dircache_create();
    return (T);
}

/* ====================================================================== */

/* dircache_add() *******************************************************
 *
 * Add single entry to directory cache (used when Prayer makes folder or
 * directory itself).
 *     dc: Directory cache
 *   name: Name to add
 *  isdir: T => Add a directory. NIL => Add a folder
 *
 * Returns: T if insertation was successful.
 ***********************************************************************/

BOOL dircache_add(struct dircache * dc, char *name, BOOL isdir)
{
    struct dircache *last, *current;
    char *next;

    if (!(name && name[0]))     /* Shouldn't be possible! */
        return (NIL);

    if (!(next = strchr(name, '/')))
        return ((dirlist_add(dc, name, isdir)) ? T : NIL);

    /* Search child nodes for next component */
    *next = '\0';
    last = NIL;
    for (current = dc->children; current; current = current->next) {
        if (!strcmp(current->name, name)) {
            *next = '/';
            return (dircache_add(current, next + 1, isdir));
        }
    }
    *next = '/';

    return (NIL);
}

/* ====================================================================== */

/* dircache_delete() *****************************************************
 *
 * Delete entry from directory cache
 *     dc: Directory cache
 *   name: Path to delete
 *
 * Returns: T if deletion was successful, NIL otherwise
 ************************************************************************/

BOOL dircache_delete(struct dircache * dc, char *name)
{
    struct dircache *last, *current;
    char *next;

    if (!(name && name[0]))     /* Shouldn't be possible! */
        return (NIL);

    if (!(next = strchr(name, '/')))
        return ((dirlist_remove(dc, name)));

    /* Search child nodes for next component */
    *next = '\0';
    last = NIL;
    for (current = dc->children; current; current = current->next) {
        if (!strcmp(current->name, name)) {
            *next = '/';
            return (dircache_delete(current, next + 1));
        }
    }
    *next = '/';

    return (NIL);
}

/* ====================================================================== */

/* dircache_rename() *****************************************************
 *
 * Rename entry in dircache
 *      dc: Directory cache
 * oldname: Original name
 * newname: New name.
 ************************************************************************/

BOOL dircache_rename(struct dircache * dc, char *oldname, char *newname)
{
    struct dirlist *dl;
    char *basename;

    if ((basename = strrchr(oldname, '/'))) {
        struct dircache *parent;

        *basename = '\0';
        parent = dircache_lookup_internal(dc, oldname);
        *basename++ = '/';

        dl = (parent) ? parent->dirlist : NIL;
    } else {
        dl = dc->dirlist;
        basename = oldname;
    }

    while (dl) {
        if (!strcmp(dl->name, basename))
            break;
        dl = dl->next;
    }

    /* Couldn't find this name */
    if (!dl)
        return (NIL);

    if (!dircache_add(dc, newname, dl->isdir))
        return (NIL);

    if (!dircache_delete(dc, oldname))
        return (NIL);

    return (T);
}
