/* $Cambridge: hermes/src/prayer/prayer/dirlist.c,v 1.1.1.1 2003/04/15 13:00:03 dpc22 Exp $ */
/************************************************
 *    Prayer - a Webmail Interface              *
 ************************************************/

/* Copyright (c) University of Cambridge 2000 - 2002 */
/* See the file NOTICE for conditions of use and distribution. */

#include "prayer_session.h"

/* Class for managing directory list for a _single_ directory. Built in
 * a tree structure using companion dircache class */

/* dirlist_create() ******************************************************
 *
 * Create a new dirlist structure in the nominated pool (typically the
 * pool assocaited with the parent dircache).
 ************************************************************************/

struct dirlist *dirlist_create(struct pool *p)
{
    struct dirlist *result = pool_alloc(p, sizeof(struct dirlist));

    result->next = NIL;
    result->name = NIL;
    result->isdir = NIL;

    return (result);
}

/* dirlist_free() ********************************************************
 *
 * Free directory list. Actally a NOOP at the moment: dircache_free()
 * does the work.
 ************************************************************************/

void dirlist_free(struct dircache *dc)
{
    return;
}

/* ====================================================================== */

/* dirlist_add() *********************************************************
 *
 * Add a single directory or folder item to a directory list.
 *     dc: Parent Directory cache
 *   name: Name of item to add
 *  isdir: T => directory. NIL => Folder
 *
 ************************************************************************/

struct dirlist *dirlist_add(struct dircache *dc, char *name, BOOL isdir)
{
    struct dirlist *dl = dc->dirlist;
    struct dirlist *new = pool_alloc(dc->pool, sizeof(struct dirlist));

    new->next = NIL;
    new->name = pool_strdup(dc->pool, name);
    new->isdir = isdir;

    if (!dl) {
        /* New list */
        dc->dirlist = new;
        return (dc->dirlist);
    }

    if (!strcmp(name, dl->name)) {
        dl->isdir = isdir;
        return (dc->dirlist);
    }

    if (strcasecmp(name, dl->name) < 0) {
        /* Insert node at start of list */
        new->next = dl;
        dc->dirlist = new;
        return (dc->dirlist);
    }

    while (dl->next) {
        if (!strcmp(name, dl->name)) {
            dl->isdir = isdir;
            return (dc->dirlist);
        }

        if (strcasecmp(name, dl->next->name) < 0) {
            /* Insert node in middle of list */
            new->next = dl->next;
            dl->next = new;
            return (dc->dirlist);
        }
        dl = dl->next;
    }

    /* Insert at end of list */
    new->next = NIL;
    dl->next = new;
    return (dc->dirlist);
}

/* ====================================================================== */

/* dirlist_remove() ******************************************************
 *
 * Remove a single entry from directory list (unlinks from relevant list,
 * however doesn't attempt to free() anything as names currently stored
 * in pool associated with parent dircache)
 *    dc: Directory Cache
 *  name: Name to remove
 ************************************************************************/

BOOL dirlist_remove(struct dircache * dc, char *name)
{
    struct dirlist *dl = dc->dirlist;
    struct dirlist *last;

    if (!dl)
        return (NIL);

    /* Remove first element from list */
    if (!strcmp(dl->name, name)) {
        dc->dirlist = dl->next;
        return (T);
    }

    /* Remove from middle/end of chain */
    last = dl;
    dl = dl->next;
    while (dl) {
        if (!strcmp(dl->name, name)) {
            last->next = dl->next;
            return (T);
        }
        last = dl;
        dl = dl->next;
    }

    /* Not found */
    return (NIL);
}

/* ====================================================================== */

/* Callback routine that c-client calls when generating a directory list */

static struct {
    struct pool *pool;
    struct dirlist *list;
    char *prefix;
    unsigned long count;
    unsigned long prefix_length;
} callback;

static void
dirlist_callback(MAILSTREAM * stream,
                 int delimiter, char *name, long attributes)
{
    struct dirlist *di, *dl =
        pool_alloc(callback.pool, sizeof(struct dirlist));

    callback.count++;

    if (callback.prefix && callback.prefix[0]) {
        if (strncmp(name, callback.prefix, callback.prefix_length) != 0)
            return;             /* Didn't match prefix */

        name += callback.prefix_length;
    }

    if (*name == '\0')
        return;                 /* Empty name */

    dl->next = NIL;
    dl->name = pool_strdup(callback.pool, name);
    dl->isdir = (attributes & LATT_NOINFERIORS) ? NIL : T;

    /* Insertion sort algorithm */

    /* Insert as first elm in empty list? */
    if (callback.list == NIL) {
        callback.list = dl;
        return;
    }

    /* Insert as first elm in non-empty list? */
    if (strcasecmp(callback.list->name, dl->name) > 0) {
        dl->next = callback.list;
        callback.list = dl;
        return;
    }

    /* Insert in middle of list? */
    for (di = callback.list; di->next; di = di->next) {
        if (strcasecmp(di->next->name, dl->name) > 0) {
            dl->next = di->next;
            di->next = dl;
            return;
        }
    }

    /* Insert at end of list */
    di->next = dl;
    dl->next = NIL;
}

/* ====================================================================== */

/* dirlist_populate() ****************************************************
 *
 * Populate directory list using c-client to send LIST to IMAP server.
 ************************************************************************/

struct dirlist *dirlist_populate(struct session *session, char *name)
{
    struct dircache *dc = session->dircache;
    MAILSTREAM *stream = session->stream;
    struct request *request = session->request;
    unsigned long offset;
    char *prefix = session_mailbox(session, request->pool, "");

    callback.list = NIL;
    callback.count = 0;
    callback.pool = dc->pool;
    mm_register_list_callback(dirlist_callback);

    callback.prefix = pool_strcat(request->pool, prefix, name);
    offset = strlen(callback.prefix);

    if (session->cwd[0]) {
        callback.prefix[offset++] = '/';
        callback.prefix[offset] = '\0';
    }

    callback.prefix_length = offset;
    if (!ml_list(session, stream, callback.prefix, "%"))
        return (NIL);

    return (callback.list);
}

/* ====================================================================== */

/* dirlist_filter_check() ************************************************
 *
 * Check if name matches filter (with wildcards). Used when generating
 * directory lists.
 *    name: Name to check
 *  filter: Filter to match against name.
 ************************************************************************/

BOOL dirlist_filter_check(char *name, char *filter)
{
    char *s, *t, *f;

    if (!(filter && filter[0]))
        return (T);

    for (s = name; *s; s++) {
        for (t = s, f = filter; *t && *f; t++, f++) {
            if (*f == '*') {
                if (dirlist_filter_check(t + 1, f + 1))
                    return (T);
            } else if ((*f != '?') && (tolower(*t) != tolower(*f)))
                break;
        }

        /* Entire filter matches */
        if (*f == '\0')
            return (T);
    }

    return (NIL);
}
