/* $Cambridge: hermes/src/prayer/prayer/session_idle.c,v 1.1.1.1 2003/04/15 13:00:03 dpc22 Exp $ */
/************************************************
 *    Prayer - a Webmail Interface              *
 ************************************************/

/* Copyright (c) University of Cambridge 2000 - 2002 */
/* See the file NOTICE for conditions of use and distribution. */

#include "prayer_session.h"

/* session_idle_background() ********************************************
 *
 * Process connection to idle port (generates a "session disconnected"
 * page).
 *  config: Prayer configuration
 *  sockfd: Socket descriptor for incoming connection
 *    port: Port number
 * use_ssl: This port was reserved for SSL connections.
 *
 * Returns: T if connection managed properly.
 ************************************************************************/

static BOOL
session_idle_background(struct config *config, int sockfd,
                        unsigned long port, BOOL use_ssl)
{
    struct iostream *stream = iostream_create(NIL, sockfd, 0);
    struct request *request = request_create(config, stream, NIL);
    struct user_agent *user_agent = user_agent_create(request->pool);
    struct buffer *b = request->write_buffer;
    struct pool *pool = request->pool;
    char *url;
    char *hostname;

    iostream_set_timeout(stream, config->http_timeout_session);

    if (use_ssl && !iostream_ssl_start_server(stream)) {
        log_debug("Invalid https connection on potr %lu", port);
        close(sockfd);
        return (NIL);
    }

    request->user_agent = user_agent;

    if (!request_parse(request)) {
        /* Badly formatted request */
        response_error(request, request->status);
        response_send(request);
    } else {
        html_common_start(config, b, "Session disconnected");
        bputs(b, "<h3>" CRLF);
        bputs(b, "Couldn't connect to session. Maybe it timed out?" CRLF);

        if (config->hostname_service)
            hostname = config->hostname_service;
        else
            hostname = config->hostname;

        if (use_ssl) {
            if (config->ssl_default_port != 443)
                url = pool_printf(pool, "https://%s:%lu",
                                  hostname, config->ssl_default_port);
            else
                url = pool_printf(pool, "https://%s", hostname);
        } else {
            if (config->http_icon_port != 80)
                url = pool_printf(pool, "http://%s:%lu",
                                  hostname, config->http_icon_port);
            else
                url = pool_printf(pool, "http://%s", hostname);
        }

        bprintf(b, ("<p>Click <a target=\"_top\" href=\"%s\">here</a>"
                    " to login again</p>" CRLF), url);
        bprintf(b, "</h3>" CRLF);
        html_common_end(b);

        request->error = T;
        response_html(request, 404);
        response_send(request);
    }

    /* Flush any data in output buffer */
    iostream_close(stream);
    request_free(request);

    return (T);
}

/* ====================================================================== */

/* session_idle() *******************************************************
 *
 * Accept and process incoming connection to idle port
 *   config: Prayer configuration
 *   sockfd: Socket descriptor for bound socket (thing to accept() on)
 *     port: Port number
 *  use_ssl: T => this port reserved for SSL connections.
 ***********************************************************************/

BOOL
session_idle(struct config * config, int sockfd,
             unsigned long port, BOOL use_ssl)
{
    int newsockfd;
    pid_t childpid;

    /* Have new incoming connection:
     * Accept in foreground to avoid race conditions */
    if ((newsockfd = os_accept_inet(sockfd, NIL)) < 0) {
        log_panic("[session_idle()] accept() failed, %s", strerror(errno));
        return (NIL);
    }

    /* Process entry as asynchronous task, return immediately */

    if ((childpid = fork()) < 0) {
        log_panic("[session_idle()] fork() failed: %s", strerror(errno));
        return (NIL);
    }

    if (childpid != 0) {
        close(newsockfd);
        return (T);             /* Parent process */
    }

    /* Child process: send out error response asynchronously */
    if (!os_signal_child_clear())
        abort();

    close(sockfd);

    session_idle_background(config, newsockfd, port, use_ssl);
    close(newsockfd);
    exit(0);
    /* NOTREACHED */
}
