/* $Cambridge: hermes/src/prayer/prayer/session_inet.c,v 1.1.1.1 2003/04/15 13:00:03 dpc22 Exp $ */
/************************************************
 *    Prayer - a Webmail Interface              *
 ************************************************/

/* Copyright (c) University of Cambridge 2000 - 2002 */
/* See the file NOTICE for conditions of use and distribution. */

#include "prayer_session.h"

/* ====================================================================== */

/* Support routines for managing lists of connections */

struct connection {
    struct connection *next;
    struct iostream *stream;
    struct request *request;
    time_t lastuse;
    unsigned long count;
    BOOL use_ssl;
    BOOL has_input;
};

static struct connection *connection_create(int sockfd, unsigned long port,
                                            BOOL use_ssl)
{
    struct connection *result = pool_alloc(NIL, sizeof(struct connection));

    result->next = NIL;

    if ((result->stream = iostream_create(NIL, sockfd, 0)) == NIL) {
        free(result);
        return (NIL);
    }

    if (use_ssl && !iostream_ssl_start_server(result->stream)) {
        log_debug("Invalid https connection on port %lu", port);
        close(sockfd);
        return (NIL);
    }


    result->request = NIL;
    result->lastuse = time(NIL);
    result->count = 0;
    result->use_ssl = use_ssl;

#if 0
    /* XXX Nonblocking removed from iostream: no longer any practical need */
    iostream_block(result->stream, NIL);
#endif
    return (result);
}

static void connection_close(struct connection *connection)
{
    BOOL use_ssl = connection->use_ssl;
    unsigned long count = connection->count;

    if (connection->stream) {
        iostream_flush(connection->stream);
        iostream_close(connection->stream);
    }

    if (connection->request)
        request_free(connection->request);

    free(connection);

    if (count != 1)
        log_debug("Closing %s connection, %lu valid requests processed",
                  (use_ssl) ? "https" : "http", count);
    else
        log_debug("Closing %s connection, 1 valid request processed",
                  (use_ssl) ? "https" : "http");
}

/* ====================================================================== */

/* session_inet_process_invalid() ***************************************
 *
 * Process connection attempt from incorrect IP address. Will run as
 * asynchronous thread to prevent denial of service attacks.
 *  session:
 *   sockfd: Socket for incoming connection
 *     port: Inet domain port
 *  use_ssl: Use SSL on this connection.
 *
 ***********************************************************************/

static BOOL
session_inet_process_invalid(struct session *session, int sockfd,
                             unsigned long port, BOOL use_ssl)
{
    struct config *config = session->config;
    struct iostream *stream = iostream_create(NIL, sockfd, 0);
    struct request *request = request_create(config, stream, NIL);
    struct buffer *b = request->write_buffer;

    iostream_set_timeout(stream, config->http_timeout_session);

    if (use_ssl && !iostream_ssl_start_server(stream)) {
        log_debug("Invalid https connection on port %lu", port);
        close(sockfd);
        return (NIL);
    }

    request->user_agent =
        user_agent_clone(request->pool, session->user_agent);

    if (!request_parse(request)) {
        /* Badly formatted request */
        response_error(request, request->status);
        response_send(request);
    } else {
        html_common_start(config, b, "Session disconnected");
        bputs(b, "<h3>" CRLF);
        bputs(b, "Couldn't connect to session. Maybe it timed out?" CRLF);
        bprintf(b,
                ("<p>Click <a target=\"_top\" href=\"%s/login/%s\">here</a>"
                 " to login again</p>" CRLF), session->url_prefix,
                request->argv[1]);
        bputs(b, "</h3>" CRLF);
        html_common_end(b);

        log_misc("Request from invalid IP address to port %lu", port);

        request->error = T;
        response_html(request, 404);
        response_send(request);
    }

    /* Flush any data in output buffer */
    iostream_close(stream);
    request_free(request);

    return (T);
}

/* ====================================================================== */

/* session_inet_process_request() ***************************************
 *
 * Collect a single HTTP request on a given connection. Dispatches request
 * when complete.
 *     session:
 *  connection: HTTP connection with pending data.
 *
 * Returns: T if connection should be left open after request complete
 *          NIL => close this connection: typically implies error
 ***********************************************************************/

static BOOL
session_inet_process_request(struct session *session,
                             struct connection *connection)
{
    struct config *config = session->config;
    struct request *request;
    BOOL persist;
    struct iostream *stream = connection->stream;

    if (connection->request == NIL)
        connection->request = request_create(config, stream, NIL);

    request = connection->request;
    request->user_agent = session->user_agent;

    if (session->telemetry)
        request_telemetry(request, session->telemetry_fd,
                          session->telemetry_all);

    if (!request_parse(request)) {
        if (!request->iseof) {
            /* Badly formatted request */
            response_error(request, request->status);
            response_send(request);
        }

        request_free(request);
        connection->request = NIL;

        /* Must flush response to client */
        iostream_flush(stream);
        return (NIL);
    }

    if (!request_complete(request))
        return (T);

    /* Have complete request: process it... */
    session->request = connection->request;

    session_exchange(session);
    persist = request->persist;

    response_send(request);
    session_accesslog(session, request);

    request_free(request);
    session->request = NIL;

    connection->request = NIL;
    connection->count++;

    /* Must flush response to client */
    iostream_flush(stream);

    return (persist);
}

/* ====================================================================== */

/* session_inet_process_input() ******************************************
 *
 * Process pending input on all active HTTP connections.
 *      session: 
 * connectionsp: Ptr to list of connections. Will be updated
 *               connections are closed.
 *      timeout: Timeout for idle connections.
 *
 ************************************************************************/

static BOOL
session_inet_process_input(struct session *session,
                           struct connection **connectionsp,
                           unsigned long timeout)
{
    struct connection *connections = *connectionsp;
    struct connection *current, *last, *next;

    last = NIL;
    current = connections;
    while (current && !session->want_disconnect) {
        BOOL close = NIL;

        if (current->has_input) {
            if (!session_inet_process_request(session, current))
                close = T;
        } else if ((timeout > 0)
                   && ((current->lastuse + timeout) < time(NIL)))
            close = T;

        next = current->next;
        if (close) {
            if (last)
                last->next = next;      /* Remove from middle/end of list */
            else
                connections = next;     /* Remove from start of list */

            connection_close(current);
        } else
            last = current;     /* Keep this link... */

        current = next;
    }

    *connectionsp = connections;
    return (T);
}

/* session_inet_close_all() **********************************************
 *
 * Close all HTTP connections. 
 *   connectionsp: Ptr to list of connections. 
 ************************************************************************/

static BOOL session_inet_close_all(struct connection **connectionsp)
{
    struct connection *current, *next;

    for (current = *connectionsp; current; current = next) {
        next = current->next;
        connection_close(current);
    }
    *connectionsp = NIL;
    return (T);
}

/* ====================================================================== */

/* session_inet_loop() ***************************************************
 *
 * Master loop for inet login session
 *   session:
 *    sockfd: Socket descriptor for this session. Thing we accept() on
 *      port: Inet port (for errors)
 *   use_ssl: Incoming connections will be SSL connections.
 ************************************************************************/

static BOOL
session_inet_loop(struct session *session, int sockfd,
                  unsigned long port, BOOL use_ssl)
{
    struct config *config = session->config;
    int newsockfd;
    fd_set readfds;
    struct timeval timeout;
    unsigned long select_timeout = config->http_timeout_session;
    int rc;
    struct connection *connections = NIL;
    struct connection *current;
    int maxfd;
    struct ipaddr ipaddr_struct, *ipaddr = &ipaddr_struct;
    BOOL idle = NIL;

    while (session->want_disconnect == NIL) {
        /* Process any outstanding requests on existing connections */
        for (current = connections; current; current = current->next)
            current->has_input =
                iostream_have_buffered_input(current->stream, T);

        session_inet_process_input(session, &connections,
                                   config->http_timeout_session);

        /* Either client or server wants out */
        if (session->want_disconnect)
            break;

        /* Requests should queue on sockfd until we are ready to serve them */
        FD_ZERO(&readfds);
        FD_SET(sockfd, &readfds);
        maxfd = sockfd;

        for (current = connections; current; current = current->next) {
            int fd = current->stream->fd;

            FD_SET(fd, &readfds);
            if (fd > maxfd)
                maxfd = fd;
        }

        if (connections)
            select_timeout = config->http_timeout_session;
        else if (idle) {
            if ((config->session_timeout_compose > 0) &&
                !strcmp(memblock_string(session->last_cmd), "compose"))
                select_timeout
                    =
                    config->session_timeout_compose -
                    config->session_idle_time;
            else if ((config->session_timeout > 0))
                select_timeout =
                    config->session_timeout - config->session_idle_time;
        } else {
            if (config->session_idle_time > 0)
                select_timeout = config->session_idle_time;
            else if ((config->session_timeout_compose > 0) &&
                     !strcmp(memblock_string(session->last_cmd),
                             "compose"))
                select_timeout = config->session_timeout_compose;
            else
                select_timeout = config->session_timeout;
        }

        if (select_timeout > 0) {
            timeout.tv_sec = select_timeout;
            timeout.tv_usec = 0;
            do {
                rc = select(maxfd + 1, &readfds, NIL, NIL, &timeout);
            }
            while ((rc < 0) && (errno == EINTR));
        } else {
            do {
                rc = select(maxfd + 1, &readfds, NIL, NIL, NIL);
            }
            while ((rc < 0) && (errno == EINTR));
        }

        if ((rc < 0) && (errno != EINTR)) {
            session_paniclog(session,
                             "[session_inet_loop()] select() failed: %s",
                             strerror(errno));
            return (NIL);
        }

        /* Reopen log files if required */
        session_log_ping(session);
        log_misc_ping();

        /* Update RSAkey every once in a while */
        iostream_check_rsakey(session->config);

        if (rc == 0) {
            if (connections == NIL) {
                if (!idle && (config->session_idle_time > 0)) {
                    session_log(session,
                                "[session_unix_loop] Session going to sleep");
                    session_streams_idle(session);
                    idle = T;
                    continue;
                }

                /* Session timeout: no activity in a long time */
                session_log(session,
                            "[session_inet_loop] Session timed out");
                break;
            } else {
                /* Connection timeout: close down all existing connections */
                session_inet_close_all(&connections);
                continue;
            }
        }

        if (idle) {
            session_log(session,
                        "[session_unix_loop] Idle session waking up");
            idle = NIL;
        }

        for (current = connections; current; current = current->next)
            current->has_input = FD_ISSET(current->stream->fd, &readfds);

        session_inet_process_input(session, &connections,
                                   config->http_timeout_session);

        /* Either client or server wants out */
        if (session->want_disconnect)
            break;

        if (!FD_ISSET(sockfd, &readfds))
            continue;

        if ((newsockfd = os_accept_inet(sockfd, ipaddr)) < 0)
            return (NIL);

        if (config->fix_client_ipaddr) {
            if (!ipaddr_compare(session->ipaddr, ipaddr)) {
                session_inet_process_invalid(session, newsockfd, port,
                                             use_ssl);
                continue;
            }
        }

        log_debug("Processing %s connection from %s on port %lu",
                  use_ssl ? "https" : "http",
                  ipaddr_text(session->ipaddr), session->session_port);

        /* Set close on exec so subprocesses can't interfere */
        if (fcntl(newsockfd, F_SETFD, FD_CLOEXEC) < 0) {
            session_paniclog(session, ("[session_inet_loop()] fcntl() "
                                       "(close-on-exec) failed, errno =%d"),
                             errno);
            return (NIL);
        }

        /* Add new connection to start of the active list */
        current = connection_create(newsockfd, port, use_ssl);

        if (current) {
            current->next = connections;
            connections = current;
        }
    }

    session_inet_close_all(&connections);
    return (T);
}

/* ====================================================================== */

/* session_inet() ********************************************************
 *
 * Run session using inet socket for communication
 *  session:
 *   stream: iostream connection to prayer frontend. Used to send
 *           URL for sucessful login request.
 *     port: Inet domain Port which corresponds to sockfd.
 *   sockfd: Socket for incoming connections: thing that we will accept() on
 *  use_ssl: This login session will use SSL connections
 *
 * Returns: T if session successful. NIL => startup error.
 ************************************************************************/

/* XXX Obsolete: eradicate */

BOOL
session_inet(struct session * session,
             struct iostream * stream,
             unsigned long port, int sockfd, BOOL use_ssl)
{
    /* Redefine session urls */
    session->use_ssl = use_ssl;
    session->session_port = port;
    session->is_direct = T;
    session_setup_urls(session);

    log_debug("New inet session: User: %s, Port: %lu",
              session->username, session->session_port);

    /* Login suceeded: okay to tell frontend where to find us now! */
    ioprintf(stream, "OK %s@init" CRLF, session->url_prefix_session);
    ioflush(stream);
    iostream_close(stream);

    setproctitle("Login session for %s", session->username);

    /* Start session loop: accept process connections */
    session_inet_loop(session, sockfd, port, use_ssl);

    log_debug("Closing inet session: User=%s, Port=%lu",
              session->username, session->session_port);

    return (T);
}
